/* 
 * File:   Bag.cpp
 * Author: michael
 * 
 * Created on 24. Oktober 2015, 11:20
 */

#include "Bag.h"

#include <boost/random/uniform_int.hpp>

using namespace boost;
using namespace boost::chrono;
using namespace rsbag;
using namespace rsc::logging;
using namespace rsc::threading;
using namespace rsb;
using namespace converter;
using namespace std;

const format Bag::URI_FORMAT = format("socket://localhost:%1%%2%");
const Scope Bag::SCOPE_PREFIX = Scope("/__rsbag/cpp/");
const Scope Bag::LISTEN_SCOPE = SCOPE_PREFIX.concat(Scope("/listen"));
const Scope Bag::CONTROL_SCOPE = SCOPE_PREFIX.concat(Scope("/control"));
const int Bag::SKIP_COMPONENTS = Bag::LISTEN_SCOPE.getComponents().size();

Bag::iterator::iterator() : logger(Logger::getLogger("rsbag.Bag.iterator")) {
}

Bag::iterator::iterator(Bag* bag, long eventNumber) : bag(bag), eventNumber(eventNumber),
logger(Logger::getLogger("rsbag.Bag.iterator")), lastFetchedEvent(-1) {
}

Bag::iterator::iterator(const iterator& other) : logger(Logger::getLogger("rsbag.Bag.iterator")) {
    this->bag = other.bag;
    this->eventNumber = other.eventNumber;
    this->event = other.event;
}

bool Bag::iterator::operator==(const iterator &b) const {
    return b.bag == this->bag && b.eventNumber == this->eventNumber;
}

bool Bag::iterator::operator!=(const iterator &b) const {
    return b.bag != this->bag || b.eventNumber != this->eventNumber;
}

rsb::EventPtr& Bag::iterator::operator*() {
    fetchEvent();
    return event;
}

rsb::EventPtr& Bag::iterator::operator->() {
    fetchEvent();
    return event;
}

void Bag::iterator::fetchEvent() {
    if (eventNumber < bag->size() && lastFetchedEvent != eventNumber) {
        event = bag->getEvent(eventNumber);
        lastFetchedEvent = eventNumber;
    }
}

Bag::iterator Bag::iterator::operator++(int) {
    iterator ret(*this);

    eventNumber++;

    return ret;
}

Bag::iterator& Bag::iterator::operator++() {
    RSCDEBUG(logger, str(format("Fetching event with number %1%.") % eventNumber));

    eventNumber++;

    return *this;
}

Bag::Bag(string bagfile, set<string> channels, int rsbag_init_time, string rsbag) :
logger(Logger::getLogger("rsbag.Bag")), timeout(5),
queue(new EventQueue(1)) {
    RSCINFO(logger, str(format("Will open bag %1% with rsbag %2%.") % bagfile % rsbag));

    // get random port for this bag.
    random::mt19937 random_generator;
    random_generator.seed(rsc::misc::currentTimeMicros());
    random::uniform_int_distribution<> distr(1025, 65535);
    port = distr(random_generator);

    // setup arguments and environment
    vector<string> arguments = buildArguments(bagfile, channels);
    vector<string> environment = buildEnvironment();

    // start rsbag with the arguments given
    bag_process = ProcessPtr(new Process(rsbag, arguments, environment));
    bag_process->start();
    
    // initial wait time
    this_thread::sleep_for(milliseconds(1000));

    // try to connect to rsbag
    bool success = false;
    for(int i=0; i<rsbag_init_time; i++) {

        try {
            // create remote server for controlled playback
            bag_server = getFactory().createRemoteServer(CONTROL_SCOPE, getParticipantConfig(), getParticipantConfig());

            // create listener and add queue handler to receive events from
            // rsbag.
            listener = getFactory().createListener(LISTEN_SCOPE, getParticipantConfig());
            listener->addHandler(HandlerPtr(new util::EventQueuePushHandler(queue)));

            // fetch the length (to make sure everything's okay)
            numEvents = *bag_server->call<long>("length", timeout);
            
            success = true;
            break;
        } catch(FutureException ex) {
            RSCINFO(logger, str(format("Try %1% of %2% failed. Will retry!") % i % rsbag_init_time));
        } catch(runtime_error ex) {
            RSCINFO(logger, str(format("Try %1% of %2% failed. Will retry!") % i % rsbag_init_time));
        }
        
        this_thread::sleep_for(milliseconds(500));
    }
    
    // if the connection failed the bag is invalid
    if(!success) {
        throw runtime_error("Startup of rsbag took too long. maybe increase rsbag_init_time?");
    }
}

ParticipantConfig Bag::getParticipantConfig() {
    ParticipantConfig config;

    // configure one socket transport with the rolled port and localhost as
    // server.
    ParticipantConfig::Transport socketTransport =
            ParticipantConfig::Transport("socket");

    socketTransport.mutableOptions().set<string>("enabled", "1");
    socketTransport.mutableOptions().set<string>("host", "localhost");
    socketTransport.mutableOptions().set<string>("port",
                                                 lexical_cast<string>(port));
    socketTransport.mutableOptions().set<string>("server", "0");

    config.addTransport(socketTransport);

    return config;
}

vector<string> Bag::buildArguments(string file, set<string> channels) {
    vector<string> arguments;

    // use play subcommand
    arguments.push_back("play");

    arguments.push_back("--show-progress=none");
    arguments.push_back("--on-error=continue");

    typedef set<string>::const_iterator ChannelIt;

    // defile channels
    for (ChannelIt it = channels.begin(); it != channels.end(); it++) {
        arguments.push_back("--channel=" + *it);
    }

    // remote controlled replay
    arguments.push_back("--replay-strategy");

    string control_uri = str(format(URI_FORMAT) % port % CONTROL_SCOPE.toString());

    arguments.push_back(str(format("remote-controlled :uri \"%1%\" :adjustments"
                                   "((:original_send (:copy :send)))")
                                   % control_uri));

    // add file to replay
    arguments.push_back(file);

    // add listen uri so only we can read the events.
    string listen_uri = str(format(URI_FORMAT) % port % LISTEN_SCOPE.toString());
    arguments.push_back(listen_uri);

    return arguments;
}

vector<string> Bag::buildEnvironment() {
    // no special environment is needed for now
    return vector<string>();
}

EventPtr Bag::getEvent(long id) {
    RSCDEBUG(logger, str(format("Fetching event number %1%.") % id));

    // cause server to emit the event we need.
    bag_server->callAsync<void>("seek", shared_ptr<long>(new long(id)));
    bag_server->callAsync<void>("emit");

    // fetch the event from our queue
    EventPtr event = queue->pop(0);

    // fix the scope.
    Scope scope("/");
    vector<string> components = event->getScope().getComponents();
    for (int i = SKIP_COMPONENTS; i < components.size(); i++) {
        scope = scope.concat(Scope(str(format("/%1%") % components[i])));
    }
    event->setScope(scope);

    return event;
}

Bag::~Bag() {
    RSCDEBUG(logger, "Deconstructing bag, asking rsbag to quit.");
    // shut down rsbag
    bag_server->call<void>("quit", timeout);
    
    // give rsbag some time for shutdown
    this_thread::sleep_for(milliseconds(2000));
}