/* 
 * File:   Bag.h
 * Author: michael
 *
 * Created on 24. Oktober 2015, 11:20
 */
#ifndef BAG_H
#define	BAG_H

#include <iterator>
#include <string>
#include <set>

#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>
#include <boost/thread.hpp>
#include <boost/chrono/duration.hpp>

#include <rsc/logging/Logger.h>
#include <rsc/threading/SynchronizedQueue.h>
#include <rsb/util/EventQueuePushHandler.h>

#include <rsb/Event.h>
#include <rsb/Factory.h>
#include <rsb/ParticipantConfig.h>
#include <rsb/patterns/RemoteServer.h>
#include <rsb/converter/Repository.h>
#include <rsb/converter/ProtocolBufferConverter.h>

#include <rsbag/exports.h>
#include <rsbag/config.h>
#include <rsbag/Process.h>

namespace rsbag {

    /**
     * Easy the registration of standard protocol buffer converters through
     * a template function.
     */
    template <class T> void registerType() {
        try {
            rsb::converter::ProtocolBufferConverter<T>* ptr = new rsb::converter::ProtocolBufferConverter<T>();
            rsb::converter::converterRepository<std::string>()->registerConverter(
                    boost::shared_ptr< rsb::converter::ProtocolBufferConverter<T> >(ptr));
        } catch (std::exception ex) {
            RSCDEBUG(rsc::logging::Logger::getLogger("rsbag.registerType"),
                    "Skipping converter registration.");
        }
    }

    /**
     * A class representing a bag file. It will generate it's own rsbag play 
     * session on construction.
     */
    class Bag {
    public:

        /**
         * An iterator over a bag to allow standard functions. It will fetch 
         * the event from its the Bag only when dereferencing.
         */
        class iterator : public std::input_iterator_tag {
            friend class Bag;

        public:

            iterator();
            iterator(const iterator& other);

            bool operator==(const iterator &b) const;
            bool operator!=(const iterator &b) const;
            
            rsb::EventPtr& operator*();
            rsb::EventPtr& operator->();
            
            iterator& operator++();
            iterator operator++(int);

        private:
            rsc::logging::LoggerPtr logger;

            void fetchEvent();

            iterator(Bag* bag, long eventNumber);

            rsb::EventPtr event;
            Bag* bag;
            unsigned long eventNumber;
            unsigned long lastFetchedEvent;
        };

        typedef rsc::threading::SynchronizedQueue<rsb::EventPtr> EventQueue;
        typedef boost::shared_ptr< EventQueue > EventQueuePtr;
        
        static const rsb::Scope SCOPE_PREFIX;
        static const rsb::Scope LISTEN_SCOPE;
        static const rsb::Scope CONTROL_SCOPE;
        static const int SKIP_COMPONENTS;

        /** 
         * Construct a new rsbag session on a random port, supplying the given
         * channels and utilising the given rsbag binary (using 'new-style'
         * rsbag calling with binary + subcommand). A new rsbag process will
         * be instantiated at a random port. After creating the process the
         * constructor waits rsbag_init_time times 0.5 seconds until the 
         * rsbag process finished its initialisation. After the rsbag_init_time
         * 'st trial the startup will fail with an std::runtime_error being
         * thrown.
         */
        Bag(std::string bagfile, std::set<std::string> channels = std::set<std::string>(),
                int rsbag_init_time = 5, std::string rsbag = rsbag::Config::getInstallPrefix() + "/bin/rsbag");

        /**
         * returns a new iterator.
         * @return the new iterator pointing to the first event.
         */
        iterator begin() {
            return iterator(this, 0);
        }

        /**
         * returns the invalid end-iterator.
         * @return the end-iterator which cannot be dereferenced.
         */
        iterator end() {
            return iterator(this, numEvents);
        }

        /**
         * Returns the number of events.
         * @return number of events within the given channels in this bag.
         */
        unsigned long size() {
            return numEvents;
        }

        /**
         * Calls seek and emit on the rsbags remote server and retrieves the
         * emited event through the queue.
         */
        rsb::EventPtr getEvent(long id);

        /**
         * Get the participant config for this bag. The config will only provide
         * one socket transport without server support, which is supposed to 
         * connect to the rsbag instance created by this Bag.
         * Warning! You will receive events prefixed with Bag::LISTEN_SCOPE .
         * 
         * @return a configuration to create rsb participants for direct 
         * listening to events from the bag file.
         */
        rsb::ParticipantConfig getParticipantConfig();

        /**
         * Call quit on the remote server and cleanup its process.
         */
        virtual ~Bag();
    private:
        rsc::logging::LoggerPtr logger;

        rsb::ListenerPtr listener;
        EventQueuePtr queue;

        rsb::patterns::RemoteServerPtr bag_server;

        static const boost::format URI_FORMAT;

        unsigned int port;
        unsigned long numEvents;
        unsigned int timeout;

        std::vector<std::string> buildEnvironment();

        std::vector<std::string> buildArguments(std::string file, std::set<std::string> channels);

        ProcessPtr bag_process;
    };

    typedef boost::shared_ptr<Bag> BagPtr;
}

#endif	/* BAG_H */

